/**
 * @file QObject.hh
 * @class QObject
 * @ingroup objects
 * @brief base class for objects handled by QEvent and QGlobalDataManager. 
 *
 * @author Marco Vignati
 *
 * Every object that is going to be handled by QEvent or QGlobalDataManager
 * must inherit from QObject. This class only contains the validity flag
 * (that says if the object has been assigned a value)
 * and the write flag (that sets wheter the QObject added to a QEvent 
 * should be written in the output files or not). Base types that inherit from this class
 * are QInt, QDouble, QFloat (templates implementations of QBaseType), QBool and QString. To store a root TObject
 * use QTObject. Specicic QObject(s) can be found in pkg/dianadata 
 * or in the @ref objects groups of this documentation. 
 *
 * Inheriting classes must declare the macro QObjectDef in the .hh file, that contains the declaration
 * of the Fill() and Duplicate() functions. The .cc file must contain the declaration of the macro 
 * QObjectImp, that implements the two functions.
 * The .cc existence is not mandatory (a QObject can be
 * a simple data container, so the .cc is useless, QBaselineData for example has no cc), Makefiles
 * will automatically generate a .cc with the needed declaration of QObjectImp.
 *
 * Inheriting classes must implement the Clear method, and optionally can implement the methods
 * Dump (that is called to print the object, e.g. cout<<object), 
 * WriteOnStream and ReadOnStream (to perform I/O with text files).
 * If the two latter methods are not implemented, the default I/O, implemented in QObjectInspector, is used.
 */

#ifndef __QOBJECT_HH_
#define __QOBJECT_HH_

#include "QDiana.hh"
#include <TObject.h>
#include "QError.hh"
#include <iostream>

#ifdef __clang__
#pragma clang diagnostic ignored "-Woverloaded-virtual"
#endif

/** 
 * @def QObjectDef(clazz,id) 
 *  macro  that must be declared in the .hh of each inheriting class (like ClassDef for TObjects) 
 */
#define QObjectDef(clazz,id) \
    public:\
    virtual bool Fill(Diana::QObject* evData) const; \
    virtual Diana::QObject* Duplicate() const; \
    ClassDef(clazz,id);


/** 
 * @def QObjectImp(claz)
 * macro that must be declared in the .cc of each inheriting class (like ClassImp for TObjects) 
 */
#define QObjectImp(clazz) \
    using Diana::QObject;\
    ClassImp(clazz)\
    bool clazz::Fill(QObject* evData) const {\
        clazz * evDatac = dynamic_cast<clazz*>(evData);\
        if(evDatac) {\
           *evDatac = *this; \
        }\
        return evDatac != NULL; \
    }\
    QObject* clazz::Duplicate() const {\
        clazz * obj = new clazz;\
        return obj;\
    }


Q_BEGIN_NAMESPACE

class QObject : public TObject {
    public:
        /* still to be migrated
           enum Type {
           LCMeasConfigMap = 3,
           GuiSession = 5,
           CurveCrawler = 7,
           };
           */
        /** @brief TObject bits */
        enum { 
            /** @brief object has been assigned a value */
            kIsValid = BIT(15),
            /** @brief  write this object onto output files, this and the bits below should go in the handle? no because of efficiency. They are stored in the fTemporaryBits member */
            kWrite = BIT(0), // 
            /** @brief  this object has been requested by somebody */
            kIsRequired = BIT(1),
            /** @brief  this object has been just added to the event */
            kIsJustAdded = BIT(2),
        };

    
        static QObject* New(const char* object);
        /** @brief default constructor */
        QObject();
        /** @brief destructor */
        virtual ~QObject() {};

        /** @brief operator= */
        QObject& operator=(const QObject& rhs);

        /** @brief make object valid */
		void Validate() {SetBit(kIsValid,true); }

        /** @brief make object non valid */
		void InValidate() { SetBit(kIsValid,false); }

        /** @brief check wheter object is valid */
		bool IsValid() const { return TestBit(kIsValid); }

        /** @brief write or not this object */
        void SetWrite ( bool write = true ) { QObjectSetBit(kWrite,write); }
        
        /** @brief check wheter write or not this object */
        bool GetWrite() const { return QObjectTestBit(kWrite); }

        /** @brief check wheter this object is required by some module */
        bool IsRequired() const { return QObjectTestBit(kIsRequired); }

        /** @brief set that this object is required  */
        void  SetIsRequired(bool required = true) {QObjectSetBit(kIsRequired,required); }

        /** @brief check wheter this object has been just added by some module */
        bool IsJustAdded() const { return QObjectTestBit(kIsJustAdded); }

        /** @brief set that this object has been just added  */
        void  SetIsJustAdded(bool added = true) {QObjectSetBit(kIsJustAdded,added); }

        /** @brief reset members to default values */
		virtual void Clear() = 0;

        /** @brief set object to non valid and calls Clear() */
		void Reset() { InValidate(); Clear(); }


        /** 
         *  @brief calls operator= of inheriting classes. Validity flag and bits are not copied. 
         *  This function is automatically declared/implemented in derived classes by QObjectDef/QObjectImp. If evData type mismatches returns false, true otherwise.
         *
         * 
         */
        virtual bool Fill(QObject* evData) const  = 0 ;

        /** @brief calls Fill(QObject*) if this QObject IsValid(). Copies Validity flag and bits */
        void FullyFill(QObject* evData) const;

        /** 
         * @brief create object of same type,
         * this function is automatically declared/implemented in derived classes
         * by QObjectDef/QObjectImp
         */
        virtual QObject* Duplicate() const = 0;

        /** @brief print content to stream (intended for screen print or log files) */
        virtual void Dump(std::ostream& o) const;

        /** @brief print content on screen */
        virtual void Print() const { Dump(std::cout); }  // *MENU*

        /** @brief print content to stream (intended for storage in text files) */
        virtual QError WriteOnStream(std::ostream& o) const;
        /** @brief fill content from stream (intended for storage in text files) */
        virtual QError ReadFromStream(std::istream& o) ;

        /** @brief copy validity and temporary bits to another QObject */
        void CopyBits(QObject* obj) const;

	virtual void Draw(Option_t* option = "") {} ;

	private:
        /** @brief set temporary bit */
        void QObjectSetBit(int bit, bool value);
        /** @brief test temporary bit */
        bool QObjectTestBit(int bit) const;
        /** @brief temporary bits are used for runtime diana settings */
        int fTemporaryBits; //!

        ClassDef(QObject,4);
};


Q_END_NAMESPACE

std::ostream& operator<<(std::ostream&s,const Diana::QObject&t);
#endif
